"""
Settings for defaults in the build pipelines for the Materials Project
"""
from typing import List

from pydantic.fields import Field

from emmet.core.provenance import Author, History
from emmet.core.settings import EmmetSettings
from emmet.core.vasp.calc_types import TaskType


class EmmetBuildSettings(EmmetSettings):
    """
    Settings for the emmet-builder module
    The default way to modify these is to modify ~/.emmet.json or set the environment variable
    EMMET_CONFIG_FILE to point to the json with emmet settings
    """

    BUILD_TAGS: List[str] = Field(
        [], description="Tags for calculations to build materials"
    )
    EXCLUDED_TAGS: List[str] = Field(
        [],
        description="Tags to exclude from materials",
    )

    DEPRECATED_TAGS: List[str] = Field(
        [], description="Tags for calculations to deprecate"
    )

    VASP_ALLOWED_VASP_TYPES: List[TaskType] = Field(
        [t.value for t in TaskType],
        description="Allowed task_types to build materials from",
    )

    DEFAULT_REFERENCE: str = Field(
        "@article{Jain2013,\nauthor = {Jain, Anubhav and Ong, Shyue Ping and "
        "Hautier, Geoffroy and Chen, Wei and Richards, William Davidson and "
        "Dacek, Stephen and Cholia, Shreyas and Gunter, Dan and Skinner, David "
        "and Ceder, Gerbrand and Persson, Kristin a.},\n"
        "doi = {10.1063/1.4812323},\nissn = {2166532X},\n"
        "journal = {APL Materials},\nnumber = {1},\npages = {011002},\n"
        "title = {{The Materials Project: A materials genome approach to "
        "accelerating materials innovation}},\n"
        "url = {http://link.aip.org/link/AMPADS/v1/i1/p011002/s1\\&Agg=doi},\n"
        "volume = {1},\nyear = {2013}\n}\n\n@misc{MaterialsProject,\n"
        "title = {{Materials Project}},\nurl = {http://www.materialsproject.org}\n}",
        description="Default bibtex citation for all provenance",
    )

    DEFAULT_AUTHOR: Author = Field(
        Author(name="Materials Project", email="feedback@materialsproject.org"),
        description="Default Author for provenance ",
    )

    DEFAULT_HISTORY: History = Field(
        History(
            name="Materials Project Optimized Structure",
            url="http://www.materialsproject.org",
        ),
        description="Default History for provenance ",
    )



from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple

from maggma.core import Builder, Store
from maggma.utils import grouper
from pymatgen.analysis.structure_matcher import OrderDisorderElementComparator
from pymatgen.core import Structure
from pymatgen.util.provenance import StructureNL

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.provenance import ProvenanceDoc
from emmet.core.utils import group_structures


class ProvenanceBuilder(Builder):
    def __init__(
        self,
        materials: Store,
        provenance: Store,
        source_snls: List[Store],
        settings: Optional[EmmetBuildSettings] = None,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Creates provenance from source SNLs and materials

        Args:
            materials: Store of materials docs to tag with SNLs
            provenance: Store to update with provenance data
            source_snls: List of locations to grab SNLs
            query : query on materials to limit search
        """
        self.materials = materials
        self.provenance = provenance
        self.source_snls = source_snls
        self.settings = EmmetBuildSettings.autoload(settings)
        self.query = query
        self.kwargs = kwargs

        materials.key = "material_id"
        provenance.key = "material_id"
        for s in source_snls:
            s.key = "snl_id"

        super().__init__(
            sources=[materials, *source_snls], targets=[provenance], **kwargs
        )

    def ensure_indicies(self):

        self.materials.ensure_index("material_id", unique=True)
        self.materials.ensure_index("formula_pretty")

        self.provenance.ensure_index("material_id", unique=True)
        self.provenance.ensure_index("formula_pretty")

        for s in self.source_snls:
            s.ensure_index("snl_id")
            s.ensure_index("formula_pretty")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:
        self.ensure_indicies()

        # Find all formulas for materials that have been updated since this
        # builder was last ran
        q = {**self.query, "property_name": ProvenanceDoc.property_name}
        updated_materials = self.provenance.newer_in(
            self.materials,
            criteria=q,
            exhaustive=True,
        )
        forms_to_update = set(
            self.materials.distinct(
                "formula_pretty", {"material_id": {"$in": updated_materials}}
            )
        )

        # Find all new SNL formulas since the builder was last run
        for source in self.source_snls:
            new_snls = self.provenance.newer_in(source)
            forms_to_update |= set(source.distinct("formula_pretty", new_snls))

        # Now reduce to the set of formulas we actually have
        forms_avail = set(self.materials.distinct("formula_pretty", self.query))
        forms_to_update = forms_to_update & forms_avail

        self.logger.info(
            f"Found {len(forms_to_update)} new/updated systems to distribute to workers "
            f"in chunks of {len(forms_to_update)/number_splits}"
        )

        for chunk in grouper(forms_to_update, number_splits):
            yield {"formula_pretty": {"$in": chunk}}

    def get_items(self) -> Tuple[List[Dict], List[Dict]]:
        """
        Gets all materials to assocaite with SNLs
        Returns:
            generator of materials and SNLs that could match
        """
        self.logger.info("Provenance Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indicies()

        # Find all formulas for materials that have been updated since this
        # builder was last ran
        q = {**self.query, "property_name": ProvenanceDoc.property_name}
        updated_materials = self.provenance.newer_in(
            self.materials,
            criteria=q,
            exhaustive=True,
        )
        forms_to_update = set(
            self.materials.distinct(
                "formula_pretty", {"material_id": {"$in": updated_materials}}
            )
        )

        # Find all new SNL formulas since the builder was last run
        for source in self.source_snls:
            new_snls = self.provenance.newer_in(source)
            forms_to_update |= set(source.distinct("formula_pretty", new_snls))

        # Now reduce to the set of formulas we actually have
        forms_avail = set(self.materials.distinct("formula_pretty", self.query))
        forms_to_update = forms_to_update & forms_avail

        self.logger.info(f"Found {len(forms_to_update)} new/updated systems to proces")

        self.total = len(forms_to_update)

        for formulas in grouper(forms_to_update, self.chunk_size):
            snls = []
            for source in self.source_snls:
                snls.extend(
                    source.query(criteria={"formula_pretty": {"$in": formulas}})
                )

            mats = list(
                self.materials.query(
                    properties=[
                        "material_id",
                        "last_updated",
                        "structure",
                        "initial_structures",
                        "formula_pretty",
                    ],
                    criteria={"formula_pretty": {"$in": formulas}},
                )
            )

            form_groups = defaultdict(list)
            for snl in snls:
                form_groups[snl["formula_pretty"]].append(snl)

            mat_groups = defaultdict(list)
            for mat in mats:
                mat_groups[mat["formula_pretty"]].append(mat)

            for formula, snl_group in form_groups.items():

                mat_group = mat_groups[formula]

                self.logger.debug(
                    f"Found {len(snl_group)} snls and {len(mat_group)} mats"
                )
                yield mat_group, snl_group

    def process_item(self, item) -> List[Dict]:
        """
        Matches SNLS and Materials
        Args:
            item (tuple): a tuple of materials and snls
        Returns:
            list(dict): a list of collected snls with material ids
        """
        mats, source_snls = item
        formula_pretty = mats[0]["formula_pretty"]
        snl_docs = list()
        self.logger.debug(f"Finding Provenance {formula_pretty}")

        # Match up SNLS with materials
        for mat in mats:
            matched_snls = list(self.match(source_snls, mat))
            if len(matched_snls) > 0:
                doc = ProvenanceDoc.from_SNLs(
                    material_id=mat["material_id"], snls=matched_snls
                )

                doc.authors.append(self.settings.DEFAULT_AUTHOR)
                doc.history.append(self.settings.DEFAULT_HISTORY)
                doc.references.append(self.settings.DEFAULT_REFERENCE)

                snl_docs.append(doc.dict())

        return snl_docs

    def match(self, snls, mat):
        """
        Finds a material doc that matches with the given snl
        Args:
            snl ([dict]): the snls list
            mat (dict): a materials doc
        Returns:
            generator of materials doc keys
        """

        m_strucs = [Structure.from_dict(mat["structure"])] + [
            Structure.from_dict(init_struc) for init_struc in mat["initial_structures"]
        ]
        snl_strucs = [StructureNL.from_dict(snl) for snl in snls]

        groups = group_structures(
            m_strucs + snl_strucs,
            ltol=self.settings.LTOL,
            stol=self.settings.STOL,
            angle_tol=self.settings.ANGLE_TOL,
            comparator=OrderDisorderElementComparator(),
        )
        matched_groups = [
            group
            for group in groups
            if any(isinstance(struc, Structure) for struc in group)
        ]
        snls = [
            struc
            for group in matched_groups
            for struc in group
            if isinstance(struc, StructureNL)
        ]

        self.logger.debug(f"Found {len(snls)} SNLs for {mat['material_id']}")
        return snls

    def update_targets(self, items):
        """
        Inserts the new SNL docs into the SNL collection
        """

        snls = list(filter(None, chain.from_iterable(items)))

        if len(snls) > 0:
            self.logger.info(f"Found {len(snls)} SNLs to update")
            self.provenance.update(snls)
        else:
            self.logger.info("No items to update")



from datetime import datetime
from itertools import chain
from typing import Dict, Iterable, Iterator, List, Optional

from maggma.builders import Builder
from maggma.stores import Store
from maggma.utils import grouper

from pymatgen.core import Structure
from emmet.builders.settings import EmmetBuildSettings
from emmet.core.utils import group_structures, jsanitize
from emmet.core.vasp.material import MaterialsDoc
from emmet.core.vasp.task import TaskDocument

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class MaterialsBuilder(Builder):
    """
    The Materials Builder matches VASP task documents by structure similarity into materials
    document. The purpose of this builder is group calculations and determine the best structure.
    All other properties are derived from other builders.

    The process is as follows:

        1.) Find all documents with the same formula
        2.) Select only task documents for the task_types we can select properties from
        3.) Aggregate task documents based on strucutre similarity
        4.) Create a MaterialDoc from the group of task documents
        5.) Validate material document

    """

    def __init__(
        self,
        tasks: Store,
        materials: Store,
        task_validation: Optional[Store] = None,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        """
        Args:
            tasks: Store of task documents
            materials: Store of materials documents to generate
            task_validation: Store for storing task validation results
            query: dictionary to limit tasks to be analyzed
            settings: EmmetSettings to use in the build process
        """

        self.tasks = tasks
        self.materials = materials
        self.task_validation = task_validation
        self.query = query if query else {}
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        sources = [tasks]
        if self.task_validation:
            sources.append(self.task_validation)
        super().__init__(sources=sources, targets=[materials], **kwargs)

    def ensure_indexes(self):
        """
        Ensures indicies on the tasks and materials collections
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_pretty")

        # Search index for materials
        self.materials.ensure_index("material_id")
        self.materials.ensure_index("last_updated")
        self.materials.ensure_index("task_ids")

        if self.task_validation:
            self.task_validation.ensure_index("task_id")
            self.task_validation.ensure_index("valid")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:
        """Prechunk the materials builder for distributed computation"""
        temp_query = dict(self.query)
        temp_query["state"] = "successful"
        if len(self.settings.BUILD_TAGS) > 0 and len(self.settings.EXCLUDED_TAGS) > 0:
            temp_query["$and"] = [
                {"tags": {"$in": self.settings.BUILD_TAGS}},
                {"tags": {"$nin": self.settings.EXCLUDED_TAGS}},
            ]
        elif len(self.settings.BUILD_TAGS) > 0:
            temp_query["tags"] = {"$in": self.settings.BUILD_TAGS}

        self.logger.info("Finding tasks to process")
        all_tasks = {
            doc[self.tasks.key]
            for doc in self.tasks.query(temp_query, [self.tasks.key])
        }
        processed_tasks = {
            t_id
            for d in self.materials.query({}, ["task_ids"])
            for t_id in d.get("task_ids", [])
        }
        to_process_tasks = all_tasks - processed_tasks
        to_process_forms = self.tasks.distinct(
            "formula_pretty", {self.tasks.key: {"$in": list(to_process_tasks)}}
        )

        for formula_chunk in grouper(to_process_forms, number_splits):
            yield {"formula_pretty": {"$in": list(formula_chunk)}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into materials documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the Materials Colection

        Returns:
            generator or list relevant tasks and materials to process into materials documents
        """

        self.logger.info("Materials builder started")
        self.logger.info(
            f"Allowed task types: {[task_type.value for task_type in self.settings.VASP_ALLOWED_VASP_TYPES]}"
        )

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime for material documents
        self.timestamp = datetime.utcnow()

        # Get all processed tasks:
        temp_query = dict(self.query)
        temp_query["state"] = "successful"
        if len(self.settings.BUILD_TAGS) > 0 and len(self.settings.EXCLUDED_TAGS) > 0:
            temp_query["$and"] = [
                {"tags": {"$in": self.settings.BUILD_TAGS}},
                {"tags": {"$nin": self.settings.EXCLUDED_TAGS}},
            ]
        elif len(self.settings.BUILD_TAGS) > 0:
            temp_query["tags"] = {"$in": self.settings.BUILD_TAGS}

        self.logger.info("Finding tasks to process")
        all_tasks = {
            doc[self.tasks.key]
            for doc in self.tasks.query(temp_query, [self.tasks.key])
        }
        processed_tasks = {
            t_id
            for d in self.materials.query({}, ["task_ids"])
            for t_id in d.get("task_ids", [])
        }
        to_process_tasks = all_tasks - processed_tasks
        to_process_forms = self.tasks.distinct(
            "formula_pretty", {self.tasks.key: {"$in": list(to_process_tasks)}}
        )
        self.logger.info(f"Found {len(to_process_tasks)} unprocessed tasks")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        if self.task_validation:
            invalid_ids = {
                doc[self.tasks.key]
                for doc in self.task_validation.query(
                    {"valid": False}, [self.task_validation.key]
                )
            }
        else:
            invalid_ids = set()

        projected_fields = [
            "last_updated",
            "completed_at",
            "task_id",
            "formula_pretty",
            "output.energy_per_atom",
            "output.energy",
            "output.structure",
            "calcs_reversed.0.input.parameters",
            "input.is_hubbard",
            "input.hubbards",
            "input.potcar_spec",
            "orig_inputs",
            "input.structure",
            "tags",
        ]

        for formula in to_process_forms:
            tasks_query = dict(temp_query)
            tasks_query["formula_pretty"] = formula
            tasks = list(
                self.tasks.query(criteria=tasks_query, properties=projected_fields)
            )
            for t in tasks:
                if t[self.tasks.key] in invalid_ids:
                    t["is_valid"] = False
                else:
                    t["is_valid"] = True

            yield tasks

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a list of materials

        Args:
            tasks [dict] : a list of task docs

        Returns:
            ([dict],list) : a list of new materials docs and a list of task_ids that were processsed
        """

        tasks = [TaskDocument(**task) for task in items]
        formula = tasks[0].formula_pretty
        task_ids = [task.task_id for task in tasks]
        self.logger.debug(f"Processing {formula} : {task_ids}")

        grouped_tasks = self.filter_and_group_tasks(tasks)
        materials = []
        for group in grouped_tasks:
            try:
                materials.append(MaterialsDoc.from_tasks(group))
            except Exception:
                failed_ids = list({t_.task_id for t_ in group})
                self.logger.warn(
                    f"No valid ids found among ids {failed_ids}. This can be the case if the required "
                    "calculation types are missing from your tasks database."
                )
        self.logger.debug(f"Produced {len(materials)} materials for {formula}")

        return jsanitize([mat.dict() for mat in materials], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding
                processed task_ids
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        for item in docs:
            item.update({"_bt": self.timestamp})

        material_ids = list({item["material_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} materials")
            self.materials.remove_docs({self.materials.key: {"$in": material_ids}})
            self.materials.update(
                docs=docs,
                key=["material_id"],
            )
        else:
            self.logger.info("No items to update")

    def filter_and_group_tasks(
        self, tasks: List[TaskDocument]
    ) -> Iterator[List[TaskDocument]]:
        """
        Groups tasks by structure matching
        """

        filtered_tasks = [
            task
            for task in tasks
            if any(
                allowed_type is task.task_type
                for allowed_type in self.settings.VASP_ALLOWED_VASP_TYPES
            )
        ]

        structures = []

        for idx, task in enumerate(filtered_tasks):
            s = task.output.structure
            s.index: int = idx  # type: ignore
            structures.append(s)

        grouped_structures = group_structures(
            structures,
            ltol=self.settings.LTOL,
            stol=self.settings.STOL,
            angle_tol=self.settings.ANGLE_TOL,
            symprec=self.settings.SYMPREC,
        )
        for group in grouped_structures:
            grouped_tasks = [filtered_tasks[struc.index] for struc in group]  # type: ignore
            yield grouped_tasks



from collections import defaultdict
from math import ceil

import numpy as np
from maggma.builders import Builder
from maggma.utils import grouper
from pymatgen.analysis.magnetism.analyzer import CollinearMagneticStructureAnalyzer
from pymatgen.core import Structure
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.symmetry.bandstructure import HighSymmKpath

from emmet.core import SETTINGS
from emmet.core.electronic_structure import ElectronicStructureDoc
from emmet.core.utils import jsanitize


class ElectronicStructureBuilder(Builder):
    def __init__(
        self,
        tasks,
        materials,
        electronic_structure,
        bandstructure_fs,
        dos_fs,
        chunk_size=10,
        query=None,
        **kwargs
    ):
        """
        Creates an electronic structure collection from a tasks collection,
        the associated band structures and density of states file store collections,
        and the materials collection.

        Individual bandstructures for each of the three conventions are generated.

        tasks (Store): Store of task documents
        materials (Store): Store of materials documents
        electronic_structure (Store): Store of electronic structure summary data documents
        bandstructure_fs (Store): Store of bandstructures
        dos_fs (Store): Store of DOS
        chunk_size (int): Chunk size to use for processing. Defaults to 10.
        query (dict): Dictionary to limit materials to be analyzed
        """

        self.tasks = tasks
        self.materials = materials
        self.electronic_structure = electronic_structure
        self.bandstructure_fs = bandstructure_fs
        self.dos_fs = dos_fs
        self.chunk_size = chunk_size
        self.query = query if query else {}

        super().__init__(
            sources=[tasks, materials, bandstructure_fs, dos_fs],
            targets=[electronic_structure],
            chunk_size=chunk_size,
            **kwargs,
        )

    def prechunk(self, number_splits: int):
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.electronic_structure.newer_in(
            self.materials, criteria=q, exhaustive=True
        )

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.materials.key: {"$in": list(split)}}}

    def get_items(self):
        """
        Gets all items to process

        Returns:
            generator or list relevant tasks and materials to process
        """

        self.logger.info("Electronic Structure Builder Started")

        q = dict(self.query)

        mat_ids = self.materials.distinct(self.materials.key, criteria=q)
        es_ids = self.electronic_structure.distinct(self.electronic_structure.key)

        mats_set = set(
            self.electronic_structure.newer_in(
                target=self.materials, criteria=q, exhaustive=True
            )
        ) | (set(mat_ids) - set(es_ids))

        mats = [mat for mat in mats_set]

        self.logger.info(
            "Processing {} materials for electronic structure".format(len(mats))
        )

        self.total = len(mats)

        for mat in mats:
            mat = self._update_materials_doc(mat)
            yield mat

    def process_item(self, mat):
        """
        Process the band structures and dos data.

        Args:
            mat (dict): material document

        Returns:
            (dict): electronic_structure document
        """

        structure = Structure.from_dict(mat["structure"])

        self.logger.info("Processing: {}".format(mat[self.materials.key]))

        dos = None
        bs = {}
        structures = {}

        for bs_type, bs_entry in mat["bandstructure"].items():
            if bs_entry.get("object", None) is not None:
                bs[bs_type] = (
                    {
                        bs_entry["task_id"]: BandStructureSymmLine.from_dict(
                            bs_entry["object"]
                        )
                    }
                    if bs_entry
                    else None
                )

                structures[bs_entry["task_id"]] = bs_entry["output_structure"]

        if mat["dos"]:
            if mat["dos"]["object"] is not None:
                self.logger.info("Processing density of states")
                dos = {
                    mat["dos"]["task_id"]: CompleteDos.from_dict(mat["dos"]["object"])
                }

                structures[mat["dos"]["task_id"]] = mat["dos"]["output_structure"]

        if bs:
            self.logger.info(
                "Processing band structure types: {}".format(
                    [bs_type for bs_type, bs_entry in bs.items() if bs_entry]
                )
            )

        if dos is None:
            doc = ElectronicStructureDoc.from_structure(
                material_id=mat[self.materials.key],
                task_id=mat["other"]["task_id"],
                structure=structure,
                band_gap=mat["other"]["band_gap"],
                cbm=mat["other"]["cbm"],
                vbm=mat["other"]["vbm"],
                efermi=mat["other"]["efermi"],
                is_gap_direct=mat["other"]["is_gap_direct"],
                is_metal=mat["other"]["is_metal"],
                magnetic_ordering=mat["other"]["magnetic_ordering"],
            )
        else:
            doc = ElectronicStructureDoc.from_bsdos(
                material_id=mat[self.materials.key],
                structures=structures,
                dos=dos,
                is_gap_direct=mat["other"]["is_gap_direct"],
                is_metal=mat["other"]["is_metal"],
                **bs,
            )

            bgap_diff = []
            for bs_type, bs_summary in doc.bandstructure:
                if bs_summary is not None:
                    bgap_diff.append(doc.band_gap - bs_summary.band_gap)

            if any(abs(gap) > 0.25 for gap in bgap_diff):
                if doc.warnings is None:
                    doc.warnings = []
                doc.warnings.append(
                    "Absolute difference between blessed band gap and at least one\
                    line-mode calculation band gap is larger than 0.25 eV."
                )

        return doc.dict()

    def update_targets(self, items):
        """
        Inserts electronic structure documents into the electronic_structure collection

        Args:
            items ([Dict]): A list of ElectronicStructureDoc dictionaries to update
        """

        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} electronic structure docs".format(len(items)))
            self.electronic_structure.update(docs=jsanitize(items, allow_bson=True))
        else:
            self.logger.info("No electronic structure docs to update")

    def _update_materials_doc(self, mat_id):
        # find bs type for each task in task_type and store each different bs object

        mat = self.materials.query_one(
            properties=[
                self.materials.key,
                "structure",
                "inputs",
                "task_types",
                self.materials.last_updated_field,
            ],
            criteria={self.materials.key: mat_id},
        )

        mat["dos"] = {}
        mat["bandstructure"] = defaultdict(dict)
        mat["other"] = {}

        bs_calcs = defaultdict(list)
        dos_calcs = []
        other_calcs = []

        for task_id in mat["task_types"].keys():

            # Handle all line-mode tasks
            if "NSCF Line" in mat["task_types"][task_id]:

                bs_type = None

                task_query = self.tasks.query_one(
                    properties=[
                        "calcs_reversed",
                        "last_updated",
                        "input.is_hubbard",
                        "orig_inputs.kpoints",
                        "input.parameters",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                fs_id = str(task_query["calcs_reversed"][0]["bandstructure_fs_id"])

                structure = Structure.from_dict(task_query["output"]["structure"])

                kpoints = task_query["orig_inputs"]["kpoints"]
                labels_dict = {
                    label: point
                    for label, point in zip(kpoints["labels"], kpoints["kpoints"])
                    if label is not None
                }

                bs_type = self._obtain_path_type(labels_dict, structure)

                if bs_type is None:

                    bs_dict = self.bandstructure_fs.query_one(
                        {self.bandstructure_fs.key: str(task_id)}
                    )

                    if bs_dict is not None:

                        bs = BandStructureSymmLine.from_dict(bs_dict["data"])

                        bs_type = self._obtain_path_type(bs.labels_dict, bs.structure)

                is_hubbard = task_query["input"]["is_hubbard"]
                nkpoints = task_query["orig_inputs"]["kpoints"]["nkpoints"]
                lu_dt = task_query["last_updated"]

                if bs_type is not None:
                    bs_calcs[bs_type].append(
                        {
                            "fs_id": fs_id,
                            "task_id": task_id,
                            "is_hubbard": int(is_hubbard),
                            "nkpoints": int(nkpoints),
                            "updated_on": lu_dt,
                            "output_structure": structure,
                        }
                    )

            # Handle uniform tasks
            if "NSCF Uniform" in mat["task_types"][task_id]:
                task_query = self.tasks.query_one(
                    properties=[
                        "calcs_reversed",
                        "last_updated",
                        "input.is_hubbard",
                        "orig_inputs.kpoints",
                        "input.parameters",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                fs_id = str(task_query["calcs_reversed"][0]["dos_fs_id"])

                is_hubbard = task_query["input"]["is_hubbard"]

                structure = Structure.from_dict(task_query["output"]["structure"])

                if task_query["orig_inputs"]["kpoints"]["generation_style"] == "Monkhorst":
                    nkpoints = np.prod(task_query["orig_inputs"]["kpoints"]["kpoints"][0], axis=0)

                else:
                    nkpoints = task_query["orig_inputs"]["kpoints"]["nkpoints"]

                nedos = task_query["input"]["parameters"]["NEDOS"]
                lu_dt = task_query["last_updated"]

                dos_calcs.append(
                    {
                        "fs_id": fs_id,
                        "task_id": task_id,
                        "is_hubbard": int(is_hubbard),
                        "nkpoints": int(nkpoints),
                        "nedos": int(nedos),
                        "updated_on": lu_dt,
                        "output_structure": structure,
                    }
                )

            # Handle static and structure opt tasks
            if "Static" or "Structure Optimization" in mat["task_types"][task_id]:
                task_query = self.tasks.query_one(
                    properties=[
                        "last_updated",
                        "input.is_hubbard",
                        "orig_inputs.kpoints",
                        "calcs_reversed",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                structure = Structure.from_dict(task_query["output"]["structure"])

                other_mag_ordering = CollinearMagneticStructureAnalyzer(
                    structure
                ).ordering

                is_hubbard = task_query["input"]["is_hubbard"]

                last_calc = task_query["calcs_reversed"][-1]

                if last_calc["input"]["kpoints"]["generation_style"] == "Monkhorst":
                    nkpoints = np.prod(
                        last_calc["input"]["kpoints"]["kpoints"][0], axis=0
                    )
                else:
                    nkpoints = last_calc["input"]["kpoints"]["nkpoints"]

                lu_dt = task_query["last_updated"]

                other_calcs.append(
                    {
                        "is_static": True
                        if "Static" in mat["task_types"][task_id]
                        else False,
                        "task_id": task_id,
                        "is_hubbard": int(is_hubbard),
                        "nkpoints": int(nkpoints),
                        "magnetic_ordering": other_mag_ordering,
                        "updated_on": lu_dt,
                        "calcs_reversed": task_query["calcs_reversed"],
                    }
                )

        updated_materials_doc = self._obtain_blessed_calculations(
            mat, bs_calcs, dos_calcs, other_calcs
        )

        return updated_materials_doc

    def _obtain_blessed_calculations(
        self, materials_doc, bs_calcs, dos_calcs, other_calcs
    ):

        bs_types = ["setyawan_curtarolo", "hinuma", "latimer_munro"]

        for bs_type in bs_types:
            # select "blessed" bs of each type
            if bs_calcs[bs_type]:
                sorted_bs_data = sorted(
                    bs_calcs[bs_type],
                    key=lambda entry: (
                        entry["is_hubbard"],
                        entry["nkpoints"],
                        entry["updated_on"],
                    ),
                    reverse=True,
                )

                materials_doc["bandstructure"][bs_type]["task_id"] = sorted_bs_data[0][
                    "task_id"
                ]

                bs_obj = self.bandstructure_fs.query_one(
                    criteria={"fs_id": sorted_bs_data[0]["fs_id"]}
                )

                materials_doc["bandstructure"][bs_type]["object"] = (
                    bs_obj["data"] if bs_obj is not None else None
                )

                materials_doc["bandstructure"][bs_type]["output_structure"] = sorted_bs_data[0]["output_structure"]

            if dos_calcs:

                sorted_dos_data = sorted(
                    dos_calcs,
                    key=lambda entry: (
                        entry["is_hubbard"],
                        entry["nkpoints"],
                        entry["nedos"],
                        entry["updated_on"],
                    ),
                    reverse=True,
                )

                materials_doc["dos"]["task_id"] = sorted_dos_data[0]["task_id"]

                dos_obj = self.dos_fs.query_one(
                    criteria={"fs_id": sorted_dos_data[0]["fs_id"]}
                )
                materials_doc["dos"]["object"] = (
                    dos_obj["data"] if dos_obj is not None else None
                )

                materials_doc["dos"]["output_structure"] = sorted_dos_data[0]["output_structure"]

            if other_calcs:

                sorted_other_data = sorted(
                    other_calcs,
                    key=lambda entry: (
                        entry["is_static"],
                        entry["is_hubbard"],
                        entry["nkpoints"],
                        entry["updated_on"],
                    ),
                    reverse=True,
                )

                materials_doc["other"]["task_id"] = str(sorted_other_data[0]["task_id"])

                task_output_data = sorted_other_data[0]["calcs_reversed"][-1]["output"]
                materials_doc["other"]["band_gap"] = task_output_data["bandgap"]
                materials_doc["other"]["magnetic_ordering"] = sorted_other_data[0][
                    "magnetic_ordering"
                ]

                materials_doc["other"]["is_metal"] = (
                    materials_doc["other"]["band_gap"] == 0.0
                )

                for prop in ["efermi", "cbm", "vbm", "is_gap_direct", "is_metal"]:

                    # First try other calcs_reversed entries if properties are not found in last
                    if prop not in task_output_data:
                        for calc in sorted_other_data[0]["calcs_reversed"]:
                            if calc["output"].get(prop, None) is not None:
                                materials_doc["other"][prop] = calc["output"][prop]
                    else:
                        materials_doc["other"][prop] = task_output_data[prop]

        return materials_doc

    @staticmethod
    def _obtain_path_type(
        labels_dict,
        structure,
        symprec=SETTINGS.SYMPREC,
        angle_tolerance=SETTINGS.ANGLE_TOL,
        atol=1e-5,
    ):

        bs_type = None

        if any([label.islower() for label in labels_dict]):
            bs_type = "latimer_munro"
        else:
            for ptype in ["setyawan_curtarolo", "hinuma"]:
                hskp = HighSymmKpath(
                    structure,
                    has_magmoms=False,
                    magmom_axis=None,
                    path_type=ptype,
                    symprec=symprec,
                    angle_tolerance=angle_tolerance,
                    atol=atol,
                )
                hs_labels_full = hskp.kpath["kpoints"]
                hs_path_uniq = set(
                    [label for segment in hskp.kpath["path"] for label in segment]
                )

                hs_labels = {
                    k: hs_labels_full[k] for k in hs_path_uniq if k in hs_path_uniq
                }

                shared_items = {
                    k: labels_dict[k]
                    for k in labels_dict
                    if k in hs_labels
                    and np.allclose(labels_dict[k], hs_labels[k], atol=1e-3)
                }

                if len(shared_items) == len(labels_dict) and len(shared_items) == len(
                    hs_labels
                ):
                    bs_type = ptype

        return bs_type



from setuptools import find_namespace_packages, setup
from setuptools_scm import get_version

version = get_version(root="..", relative_to=__file__, version_scheme="post-release")
version = version.split(".post")[0]


setup(
    name="emmet-builders",
    use_scm_version={"root": "..", "relative_to": __file__},
    setup_requires=["setuptools_scm"],
    description="Builders for the Emmet Library",
    author="The Materials Project",
    author_email="feedback@materialsproject.org",
    url="https://github.com/materialsproject/emmet",
    packages=find_namespace_packages(include=["emmet.*"]),
    install_requires=[
        f"emmet-core~={version}",
        "maggma~=0.27.0",
    ],
    license="modified BSD",
    zip_safe=False,
)
